N              [0-9]
O              ({N}{1,3})

C              [0-9A-Fa-f]
H              ({C}{1,4})

    #include <sys/types.h>
    #include <sys/socket.h>
    #include <sys/time.h>
    #include <sys/stat.h>

    #include <netinet/in.h>

    #include <arpa/inet.h>
    #include <arpa/nameser.h>

    #include <ctype.h>
    #include <errno.h>
    #ifdef HAVE_MEMORY_H
    #include <memory.h>
    #endif
    #include <netdb.h>
    #include <resolv.h>
    #include <setjmp.h>
    #include <signal.h>
    #include <stdio.h>
    #include <stdlib.h>
    #include <string.h>
    #include <unistd.h>

    #include "gnuc.h"
    #ifdef HAVE_OS_PROTO_H
    #include "os-proto.h"
    #endif

    #include "nb_dns.h"
    #include "setsignal.h"
    #include "version.h"

    #undef yywrap
    #ifdef FLEX_SCANNER
    #define YY_NO_INPUT
    #define YY_NO_UNPUT
    #endif

    /* -x: Only print the ip addresses, one per line; -xx: include cidr width */
    #define MATCH(isv6) if (xflag > 1) { \
	    echo(); \
	    BEGIN(WIDTH); \
    } else if (xflag) { \
	    echo(); \
	    putchar('\n'); \
    } else { \
	    convert(yytext, (isv6)); \
    }
    #define NOMATCH if (!xflag) { \
	    echo(); \
    }

    int doinfo;
    int linemode;		/* convert at most one entry per line */
    #ifdef HAVE_ASYNC_DNS
    int lookup_pass;		/* true if lookup only */
    #endif
    int triedone;
    int xflag;			/* only print matches with no translation */

    void convert(const char *, int);
    void reportinfo(void);
    int yywrap(void);
    int yylex(void);

    static inline void
    echo()
    {
    #ifdef HAVE_ASYNC_DNS
	    if (!lookup_pass)
    #endif
		    (void)fwrite(yytext, yyleng, 1, yyout);
    #ifdef SIGINFO
	    if (doinfo) {
		    reportinfo();
		    doinfo = 0;
	    }
    #endif
    }
    #define ECHO echo()

%x WIDTH

%%

::{O}(\.{O}){3}		MATCH(1);
{O}(\.{O}){3}		MATCH(0);

:(:{H}){1,7}		MATCH(1);
{H}(:{H}){7}		MATCH(1);
{H}:(:{H}){1,6}		MATCH(1);
({H}:){2}(:{H}){1,5}	MATCH(1);
({H}:){3}(:{H}){1,4}	MATCH(1);
({H}:){4}(:{H}){1,3}	MATCH(1);
({H}:){5}(:{H}){1,2}	MATCH(1);
({H}:){6}:{H}		MATCH(1);

<WIDTH>"/"{N}+		{
			/* Output width part of cidr and newline */
			echo();
			putchar('\n');
			BEGIN(0);
			}

<WIDTH>.		{
			/* No cdir width; output newline */
			putchar('\n');
			BEGIN(0);
			}

({O}\.){1,3}		NOMATCH;	/* anti-backtrack */
{O}((\.{O}){1,2})	NOMATCH;	/* anti-backtrack */

    /*
     * Speedups: Avoid handling unmatched characters one at a time
     * via the default echo rule.
     */

{H}+			NOMATCH;
[^0-9A-Fa-f\n]+		NOMATCH;

[^0-9A-Fa-f\n]+\n	{
			NOMATCH;
			triedone = 0;
			}

<WIDTH>\n		{
			/* No cdir width; output newline */
			putchar('\n');
			triedone = 0;
			BEGIN(0);
			}

\n			{
			NOMATCH;
			triedone = 0;
			}

%%

/*
 * Copyright (c) 1989, 1990, 1991, 1992, 1993, 1994, 1996, 1998, 1999, 2000, 2001, 2002, 2004, 2005, 2009, 2013, 2014, 2016, 2017, 2020, 2023
 * The Regents of the University of California. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University nor the names of its contributors
 *       may be used to endorse or promote products derived from this software
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] __attribute__ ((unused)) =
    "@(#) Copyright (c) 1989, 1990, 1991, 1992, 1993, 1994, 1996, 1998, 1999, 2000, 2001, 2002, 2004, 2005, 2009, 2013, 2014, 2016, 2017, 2020, 2023\n\
The Regents of the University of California.  All rights reserved.\n";
static const char rcsid[] __attribute__ ((unused)) =
    "@(#) $Id: hf.l 223 2023-09-06 23:13:16Z leres $ (LBL)";
#endif

#define HSIZE 8192		/* must be a power of two */

struct htable {
	char addr[NS_IN6ADDRSZ];
	int af;			/* address family */
	int alen;
	int state;
	char *name;		/* overloaded variable */
	struct htable *next;
} htable[HSIZE];

#define STATE_FREE	0	/* not in use */
#define STATE_RAW	1	/* couldn't translate */
#define STATE_SENTPTR	2
#define STATE_HAVEPTR	3
#define STATE_SENTA	4
#define STATE_HAVEA	5

#define PLURAL(n) ((n) == 1 ? "" : "s")

int strip = 1;			/* strip local domain when possible */
int lcase = 1;			/* force lowercase */
int shortdomain;		/* strip entire domain */
const char defaultefmt[] = "%h";
const char *efmt = defaultefmt;	/* expansion format */
#ifdef HAVE_ASYNC_DNS
int asyncdns;			/* 2 pass async dns hack */
int numasync;			/* number of outstanding async requests */
int asyncfd;
#endif
int networknumber;		/* convert to network numbers */
int check;			/* check PTR's against A records */
int debug;

int tmo;			/* seconds to wait for a answer from the dns */
int doingdns;			/* true if we're waiting for the nameserver */
sigjmp_buf alrmenv;		/* siglongjmp() buffer */

/* Rate limit */
float queriespersec = 1000.0;	/* maximum queries per second with async */
struct timeval startts;		/* start time for rate limit calculation */
long querycount;		/* number of queries so far */

long cachehits;			/* number of successful cache hits */
int filecount;			/* number of files so far */
char *prog;

char domain[64];		/* current domain name (including '.') */
int domainlen;			/* length of domain name */

char azero[NS_IN6ADDRSZ];

#ifdef HAVE_ASYNC_DNS
struct nb_dns_info *nd;
#endif

int targc;
char **targv;

extern char *optarg;
extern int optind, opterr;

/* ANSI C defines this */
#ifndef __STDC__
extern char *malloc();
#endif

/* Forwards */
char *a2h(const char *, int, int);
char *addr2host(const char *, int, int);
#ifdef HAVE_ASYNC_DNS
void asyncreap(int);
#endif
float calculaterate(void);
struct htable *cacheaddr(const char *, int, int, int, const char *);
void dump(void);
const char *format(const char *, const char *, const char *);
int getdomain(void);
void gettime(struct timeval *);
struct htable *hash(const char *, int);
#ifdef SIGINFO
void info(int);
#endif
#ifdef HAVE_ASYNC_DNS
int ispipe(FILE *);
#endif
struct htable *lookupaddr(const char *, int, int);
int main(int, char **);
void massagename(char *);
void mysleep(int);
void ratelimit(void);
void timeout(int);
void usage(void) __attribute__((noreturn));

int
main(argc, argv)
	int argc;
	char **argv;
{
	char *cp, *ep;
	int op, nflags;
	long v;
#ifdef HAVE_ASYNC_DNS
	char errstr[NB_DNS_ERRSIZE];
#endif

	if (argv[0] == NULL)
		prog = "hf";
	else if ((cp = strrchr(argv[0], '/')) != NULL)
		prog = cp + 1;
	else
		prog = argv[0];

	opterr = 0;
	nflags = 0;
	while ((op = getopt(argc, argv, "1abcdf:ilnNR:t:x")) != EOF)
		switch (op) {

		case '1':
			++linemode;
			++nflags;
			break;

		case 'a':
#ifdef HAVE_ASYNC_DNS
			++asyncdns;
#else
			fprintf(stderr,
			    "%s: warning: -a not supported; ignored\n", prog);
#endif
			++nflags;
			break;

		case 'b':
			efmt = "%h(%i)";
			++nflags;
			break;

		case 'c':
			++check;
			++nflags;
			break;

		case 'd':
			++debug;
			++nflags;
			break;

		case 'f':
			if (*optarg == '\0')
				efmt = defaultefmt;
			else
				efmt = optarg;
			++nflags;
			break;

		case 'i':
			lcase = 0;
			++nflags;
			break;

		case 'l':
			strip = 0;
			++nflags;
			break;

		case 'n':
#ifdef notdef
			++networknumber;
#else
			fprintf(stderr, "%s: -n not currently impemented\n",
			    prog);
			exit(1);
#endif
			++nflags;
			break;

		case 'N':
			++shortdomain;
			++nflags;
			break;

		case 'R':
			v = strtol(optarg, &ep, 10);
			if (*ep != '\0' || v > 1000000)
				usage();
			queriespersec = (float)v;
			break;

		case 't':
			tmo = atoi(optarg);
			if (tmo <= 0)
				usage();
			++nflags;
			break;

		case 'x':
			++xflag;
			break;

		default:
			usage();
		}

	if (xflag && nflags)
		usage();

#ifdef HAVE_ASYNC_DNS
	if (asyncdns) {
		nd = nb_dns_init(errstr);
		if (nd == NULL) {
			fprintf(stderr, "%s: nb_dns_init: %s\n", prog, errstr);
			exit(1);
		}
		asyncfd = nb_dns_fd(nd);
		/* If no explicit timeout, use resolver retransmit */
		if (tmo == 0)
			tmo = _res.retrans;
	}
#endif

	/* Figure out our domain, if necessary */
	if (!strip || shortdomain || !getdomain())
		domain[0] = '\0';

	/* Up number of retries, we really want answers */
	_res.retry = 20;

	/* Don't search, we'll use only FQDNs */
	_res.options &= ~RES_DNSRCH;

	/* Use a persistent TCP connection */
	_res.options |= (RES_USEVC | RES_STAYOPEN);

	/* Start time for rate limit calculations */
	gettime(&startts);

#ifdef SIGINFO
	(void)setsignal(SIGINFO, info);
#endif

	/* Setup alarm catcher if -t */
#ifdef HAVE_ASYNC_DNS
	if (!asyncdns)
#endif
		if (tmo > 0)
			(void)setsignal(SIGALRM, timeout);

	/* Let yywrap() figure out if there are any arguments to open */
	targc = argc - optind;
	targv = &argv[optind];
	yyin = NULL;
	(void)yywrap();

	/* Process file opened by yywrap() or stdin if no arguments */
	if (yyin) {
#ifdef HAVE_ASYNC_DNS
		/* XXX depends on the type of stdin */
		/* XXX can we do a test rewind? */
#ifdef notdef
		if (asyncdns && yyin == stdin)
			fprintf(stderr,
			    "%s: warning: can't use -a on stdin\n", prog);
#endif
#endif
		yylex();
	}

	if (debug) {
		fflush(stdout);
		reportinfo();
		if (debug > 1)
			dump();
	}
	exit(0);
}

int
yywrap()
{
	char *file;
	static int didany = 0;

	/* Close file, if necessary */
	if (yyin) {
#ifdef HAVE_ASYNC_DNS
		if (asyncdns) {
			if (lookup_pass) {
				if (fseek(yyin, 0L, SEEK_SET) < 0) {
					fprintf(stderr,
					    "%s: fseek/rewind: %s\n",
					    prog, strerror(errno));
					exit(1);
				}
				yyrestart(yyin);
				lookup_pass = 0;
				asyncreap(1);
				return (0);
			}
			numasync = 0;
		}
#endif
		if (yyin != stdin)
			(void)fclose(yyin);
		yyin = NULL;
	}

	/* Spin through arguments until we run out or successfully open one */
	while (targc > 0) {
		file = targv[0];
		--targc;
		++targv;
		++didany;
		if ((yyin = fopen(file, "r")) != NULL) {
#ifdef HAVE_ASYNC_DNS
			if (asyncdns)
				lookup_pass = 1;
#endif
			++filecount;
			return (0);
		}
		perror(file);
	}
	if (!didany) {
		++didany;
		++filecount;
		yyin = stdin;
#ifdef HAVE_ASYNC_DNS
		if (asyncdns) {
			if (ispipe(yyin)) {
				fprintf(stderr,
				    "%s: warning: can't use -a on a pipe\n",
				    prog);
				asyncdns = 0;
			} else
				lookup_pass = 1;
		}
#endif
		return (0);
	}
	return (1);
}

const char *
format(const char *fmt, const char *hnstr, const char *ipstr)
{
	char *cp;
	const char *cp2, *item;
	size_t len, ilen, llen, dlen, nlen, hlen, itemlen;
	static char buf[1024];

	/* ip address length */
	ilen = strlen(ipstr);

	/* Normally we are passed the long hostname */
	if (hnstr != NULL) {
		/* Long hostname length */
		llen = strlen(hnstr);

		/* Local domain truncated hostname length */
		dlen = llen;
		if (*domain != '\0') {
			cp2 = hnstr + (llen - domainlen);
			if (cp2 > hnstr && strcasecmp(cp2, domain) == 0)
				dlen = cp2 - hnstr;
		}

		/* Local domain truncated hostname length */
		cp = strchr(hnstr, '.');
		if (cp == NULL)
			nlen = llen;
		else
			nlen = cp - hnstr;

		/* Default hostname length */
		if (shortdomain)
			hlen = nlen;
		else
			hlen = dlen;
	} else {
		/* We only have the ip address string */
		hnstr = ipstr;
		llen = strlen(hnstr);
		dlen = llen;
		nlen = llen;
		hlen = llen;
	}

	cp = buf;
	/* Leave room for EOS */
	len = sizeof(buf) - 1;
	cp2 = fmt;
	while (*cp2 != '\0' && len > 0) {
		if (*cp2 != '%') {
			*cp++ = *cp2++;
			--len;
			continue;
		}
		++cp2;
		switch (*cp2) {

		case 'h':
			/* Default hostname */
			item = hnstr;
			itemlen = hlen;
			break;

		case 'D':
			/* Local domain truncated hostname */
			item = hnstr;
			itemlen = dlen;
			break;

		case 'N':
			/* Domain truncated hostname */
			item = hnstr;
			itemlen = nlen;
			break;

		case 'i':
			/* ip address */
			item = ipstr;
			itemlen = ilen;
			break;

		case 'l':
			/* Long hostname */
			item = hnstr;
			itemlen = llen;
			break;

		default:
			/* Random character */
			*cp++ = *cp2++;
			--len;
			continue;
		}
		if (itemlen > len)
			itemlen = len;
		strncpy(cp, item, itemlen);
		cp += itemlen;
		len -= itemlen;
		++cp2;
	}
	*cp = '\0';

	return (buf);
}

int
getdomain()
{
	char *cp;
	struct hostent *hp;
	char host[128];

	if (gethostname(host, sizeof(host) - 1) < 0)
		return (0);
	if ((cp = strchr(host, '.')) == NULL) {
		/* Not already canonical */
		if (tmo > 0)
			alarm(tmo);
		doingdns = 1;
		if (sigsetjmp(alrmenv, 1))
			return (0);
		hp = gethostbyname(host);
		doingdns = 0;
		if (hp == NULL)
			return (0);
		if ((cp = strchr(hp->h_name, '.')) == NULL)
			return (0);
	}
	(void)strncpy(domain, cp, sizeof(domain));
	domain[sizeof(domain) - 1] = '\0';
	if (lcase)
		for (cp = domain; *cp; ++cp)
			if (isupper((int)*cp))
				*cp = tolower(*cp);
	domainlen = strlen(domain);
	return (1);
}

void
gettime(struct timeval *tvp)
{
	if (gettimeofday(tvp, NULL) < 0) {
		fprintf(stderr, "%s: gettimeofday: %s\n",
		    prog, strerror(errno));
		exit(1);
	}
}

void
ratelimit()
{
	if (queriespersec > 0.0) {
		if (querycount > 0)
			while (calculaterate() > queriespersec)
				mysleep(10);
		++querycount;
	}
}

void
timeout(int signo)
{
	if (doingdns) {
		doingdns = 0;
		siglongjmp(alrmenv, 1);
	}
}

/* Convert address to hostname via the dns */
char *
a2h(const char *ap, int alen, int af)
{
	char **pp, *host2;
	size_t len;
	static struct hostent *hp;
	static char *host = NULL;
	static size_t hostlen = 0;

	/* Look up the PTR */
	if (tmo > 0)
		alarm(tmo);
	doingdns = 1;
	if (sigsetjmp(alrmenv, 1))
		return (NULL);

	/* Handle rate limiting */
	ratelimit();

	hp = gethostbyaddr(ap, alen, af);
	doingdns = 0;
	if (hp == NULL)
		return (NULL);

	len = strlen(hp->h_name) + 1;
	if (hostlen < len) {
		if (len < 132)
			len = 132;
		if (host == NULL)
			host = malloc(len);
		else {
			host2 = realloc(host, len);
			if (host2 != NULL)
				host = host2;
			else {
				free(host);
				host = NULL;
			}
		}
		if (host == NULL) {
			hostlen = 0;
			return (NULL);
		}
		hostlen = len;
	}
	(void)strcpy(host, hp->h_name);

	/* Done if we aren't checking */
	if (!check)
		return (host);

#ifndef HAVE_GETHOSTBYNAME2
	if (af != AF_INET)
		return (NULL);
#endif

	/* Check PTR against the A record */
	if (tmo > 0)
		alarm(tmo);
	doingdns = 1;
	if (sigsetjmp(alrmenv, 1))
		return (NULL);

	/* Handle rate limiting */
	ratelimit();

#ifdef HAVE_GETHOSTBYNAME2
	hp = gethostbyname2(host, af);
#else
	hp = gethostbyname(host);
#endif
	doingdns = 0;
	if (hp == NULL)
		return (NULL);
	if (af != hp->h_addrtype)
		return (NULL);

	/* Spin through ip addresses looking for a match */
	for (pp = hp->h_addr_list; *pp != NULL; ++pp)
		if (memcmp(ap, *pp, alen) == 0)
			return (host);

	return (NULL);
}

/* Convert address to hostname via the cache and/or dns */
char *
addr2host(const char *ap, int alen, int af)
{
	int state;
	char *host;
	struct htable *p;

	/* First look in hash table */
	p = lookupaddr(ap, alen, af);
	if (p != NULL)
		return (p->name);

	/* Lookup this host */
	host = a2h(ap, alen, af);
	state = STATE_RAW;
	if (host != NULL) {
		if (check)
			state = STATE_HAVEA;
		else
			state = STATE_HAVEPTR;
		massagename(host);
	}

	p = cacheaddr(ap, state, alen, af, host);
	if (p != NULL)
		return (p->name);

	return (host);
}

/* Look hash table entry for address */
struct htable *
lookupaddr(const char *ap, int alen, int af)
{
	struct htable *p;

	for (p = hash(ap, af); p != NULL; p = p->next) {
		if (p->af == af && memcmp(p->addr, ap, alen) == 0) {
#ifdef HAVE_ASYNC_DNS
			if (!asyncdns || lookup_pass)
#endif
				++cachehits;
			return (p);
		}
	}
	return (NULL);
}

void
massagename(char *name)
{
	char *cp;

	if (lcase)
		for (cp = name; *cp; ++cp)
			if (isupper((int)*cp))
				*cp = tolower(*cp);
}

void
mysleep(int ms)
{
	struct timeval tv;
	div_t res;

	res = div(ms, 1000);
	tv.tv_sec = res.quot;
	tv.tv_usec = res.rem * 1000;
	if (select(0, NULL, NULL, NULL, &tv) < 0 && errno != EINTR) {
		fprintf(stderr, "%s: mysleep: select: %s\n",
		    prog, strerror(errno));
		exit(1);
	}
}

struct htable *
cacheaddr(const char *ap, int state, int alen, int af, const char *host)
{
	struct htable *p, *p2;

	/* Don't cache zero */
	if (memcmp(ap, azero, alen) == 0)
		return (NULL);

	/* Look for existing slot in hash table */
	for (p = hash(ap, af); p != NULL; p = p->next)
		if (p->state != STATE_FREE &&
		    p->af == af &&
		    memcmp(p->addr, ap, alen) == 0)
			break;

	/* Allocate a new slot */
	if (p == NULL) {
		p = hash(ap, af);
		if (p->state != STATE_FREE) {
			/* Handle the collision */
			p2 = (struct htable *)malloc(sizeof(struct htable));
			/* Lose, lose */
			if (p2 == NULL)
				return (NULL);
			memset((char *)p2, 0, sizeof(struct htable));
			p2->next = p->next;
			p->next = p2;
			p = p2;
		}
	}

	/* Install new host */
	memmove(p->addr, ap, alen);
	p->alen = alen;
	p->af = af;
	if (host != NULL)
		p->name = strdup(host);
	if (state != 0)
		p->state = state;
	if (p->state == STATE_FREE)
		abort();

	/* Return answer entry */
	return (p);
}

void
dump()
{
	char *cp;
	int i, j, n, d;
	struct htable *p, *p2;
	char buf[132];

	d = n = 0;
	for (p = htable, i = 0; i < HSIZE; ++p, ++i)
		if (p->name) {
			++n;
			j = 0;
			for (p2 = p; p2; p2 = p2->next) {
				if ((cp = p2->name) == NULL)
					cp = "<nil>";
				else if (cp == (char *)1)
					cp = "<raw>";
				(void)fprintf(stderr, "%4d:%d ", i, j);
				if (inet_ntop(p2->af, p2->addr,
				    buf, sizeof(buf)) == NULL)
					(void)fprintf(stderr, "?");
				else
					(void)fprintf(stderr, "%s", buf);
				switch (p2->state) {

				case STATE_HAVEA:
					(void)fprintf(stderr, " HAVEA");
					break;

				case STATE_HAVEPTR:
					(void)fprintf(stderr, " HAVEPTR");
					break;

				case STATE_SENTPTR:
					(void)fprintf(stderr, " SENTPTR");
					break;

				case STATE_RAW:
					(void)fprintf(stderr, " RAW");
					break;

				default:
					(void)fprintf(stderr, " #%d",
					    p2->state);
					break;
				}
				(void)fprintf(stderr, " \"%s\"\n", cp);
				++d;
				++j;
			}
		}
	d -= n;
	(void)fprintf(stderr, "%d entries (%d dynamically linked)\n", n, d);
}

#ifdef HAVE_ASYNC_DNS
void
asyncreap(int ateof)
{
	char *host;
	int n;
	char **pp;
	struct htable *p;
	struct nb_dns_result *nr;
	struct hostent *hp;
	fd_set fds;
	struct timeval to;
	char errstr[NB_DNS_ERRSIZE];
	struct nb_dns_result xxxnr;

	nr = &xxxnr;
	memset(nr, 0, sizeof(*nr));
	while (numasync > 0) {
		FD_ZERO(&fds);
		FD_SET(asyncfd, &fds);
		/* If we're not at EOF, just poll */
		if (!ateof) {
			to.tv_sec = 0;
			to.tv_usec = 0;
		} else {
			to.tv_sec = tmo;
			to.tv_usec = 0;
		}
		n = select(asyncfd + 1, &fds, NULL, NULL, &to);
		if (n < 0) {
			fprintf(stderr, "%s: select: %s\n",
			    prog, strerror(errno));
			exit(1);
		}

		/* Done if timed out */
		if (n == 0)
			break;

		n = nb_dns_activity(nd, nr, errstr);
		if (n < 0) {
			fprintf(stderr, "%s: nb_dns_activity: %s\n",
			    prog, errstr);
			exit(1);
		}

		/* Bail if reply doesn't match any current queries */
		if (n == 0)
			continue;

		/* Decrement outstanding request counter */
		--numasync;

		/* Bail if not a good answer */
		if (nr->host_errno != NETDB_SUCCESS)
			continue;

		/* Bail if no hostname (probably shouldn't happen) */
		hp = nr->hostent;
		host = hp->h_name;
		if (host == NULL)
			continue;

		/* Recover hash table pointer */
		p = (struct htable *)nr->cookie;

		switch (p->state) {

		case STATE_SENTPTR:
			/* Are we done? */
			if (!check) {
				p->state = STATE_HAVEPTR;
				break;
			}

			/* Now look up the A record */
			if (nb_dns_host_request2(nd, host, p->af,
			    (void *)p, errstr) < 0) {
				fprintf(stderr, "%s: nb_dns_host_request: %s\n",
				    prog, errstr);
				p->state = STATE_RAW;
				free(p->name);
				p->name = NULL;
				break;
			}

			/* Cache the fact that we're looking */
			++numasync;
			p->state = STATE_SENTA;
			break;

		case STATE_SENTA:
			/* Check A against our address */
			if (p->af != hp->h_addrtype) {
				p->state = STATE_RAW;
				free(p->name);
				p->name = NULL;
				break;
			}

			/* Spin through ip addresses looking for a match */
			for (pp = hp->h_addr_list; *pp != NULL; ++pp)
				if (memcmp(p->addr, *pp, p->alen) == 0)
					break;

			if (pp == NULL) {
				p->state = STATE_RAW;
				free(p->name);
				p->name = NULL;
				break;
			}
			p->state = STATE_HAVEA;
			break;

		default:
			abort();
		}
		massagename(host);
		if (p->name != NULL)
			abort();
		if (host != NULL)
			p->name = strdup(host);
	}
}
#endif

float
calculaterate()
{
	struct timeval now, delta;

	gettime(&now);
	timersub(&now, &startts, &delta);
	return ((float)querycount) /
	    (((float)delta.tv_sec) + (((float)delta.tv_usec) / 1000000.0));
}

void
convert(const char *ipstr, int isv6)
{
	const char *hnstr;
	int alen;
	int af;
#ifdef HAVE_ASYNC_DNS
	struct htable *p;
	char errstr[NB_DNS_ERRSIZE];
	static int num = 0;
#endif
	char addr[NS_IN6ADDRSZ];

	if (isv6) {
#ifdef AF_INET6
		af = AF_INET6;
		alen = NS_IN6ADDRSZ;
#else
#ifdef HAVE_ASYNC_DNS
		if (!asyncdns || !lookup_pass)
#endif
			fputs(ipstr, stdout);
		return;
#endif
	} else {
		af = AF_INET;
		alen = NS_INADDRSZ;
	}

#ifdef HAVE_ASYNC_DNS
	if (asyncdns && lookup_pass) {
		if (inet_pton(af, ipstr, addr) != 1)
			return;

		/* Done if already in hash table */
		if (lookupaddr(addr, alen, af) != NULL)
			return;

		p = cacheaddr(addr, STATE_SENTPTR, alen, af, NULL);
		if (p == NULL)
			return;

		/* Handle rate limiting */
		ratelimit();

		if (nb_dns_addr_request2(nd, addr, af,
		    (void *)p, errstr) >= 0) {
			/* Cache the fact that we're looking */
			++numasync;
			++num;
		} else
			fprintf(stderr, "%s: nb_dns_host_request: %s\n",
			    prog, errstr);
		/* reap replies after we send a number of queries */
		if (num > 10) {
			asyncreap(0);
			num = 0;
		}
		return;
	}
#endif

	/* Only attempt to translate one address per line */
	if (linemode && triedone) {
		fputs(ipstr, stdout);
		return;
	}
	++triedone;

	if (inet_pton(af, ipstr, addr) == 1)
		hnstr = addr2host(addr, alen, af);
	else
		hnstr = NULL;
	fputs(format(efmt, hnstr, ipstr), stdout);
}

struct htable *
hash(const char *ap, int af)
{
	u_int32_t h;

	switch (af) {

	case AF_INET:
		memmove(&h, ap, sizeof(h));
		break;

	case AF_INET6:
		memmove(&h, ap + NS_IN6ADDRSZ - sizeof(h), sizeof(h));
		break;

	default:
		abort();
	}
	return (&htable[h & (HSIZE - 1)]);
}

#ifdef SIGINFO
void
info(int signo)
{
	doinfo = 1;
}
#endif

#ifdef HAVE_ASYNC_DNS
int
ispipe(FILE *f)
{
	struct stat sbuf;

	if (fstat(fileno(f), &sbuf) < 0) {
		fprintf(stderr, "%s: fstat: %s\n", prog, strerror(errno));
		exit(1);
	}
	if ((sbuf.st_mode & S_IFMT) != S_IFREG)
		return (1);
	return (0);
}
#endif

void
reportinfo(void)
{
	struct timeval now, delta;

	gettime(&now);
	timersub(&now, &startts, &delta);
	(void)fprintf(stderr, "[%d file%s, %ld querie%s,"
	    " %ld.%03ld seconds, %ld cache hit%s, %.1f queries/sec]\n",
	    filecount, PLURAL(filecount),
	    querycount, PLURAL(querycount),
	    delta.tv_sec, delta.tv_usec / 100,
	    cachehits, PLURAL(cachehits),
	    calculaterate());
}

void
usage()
{

	(void)fprintf(stderr, "Version %s\n", version);
	(void)fprintf(stderr, "usage: %s [-1abcdilN] [-f format] [-R qps]"
	    " [-t secs] [file ...]\n", prog);
	(void)fprintf(stderr, "       %s -x [file ...]\n", prog);
	exit(1);
}
